int N = 40;
float W = 0.25;
float H = 0.5;
float E = 50.0;
float EL = 10.0;
float CEL = -0.25;
float SW;
float SH;
float SE;
float SEL;
float TX;
float TY;

float THETA = 0.523598776;
float A = cos(THETA);
float B = sin(THETA);

boolean scrolling;
boolean painting;
float scrolltx;
float scrollty;

Tunit lastHeroicTunit;

Tunit[] tunits;

class Path {
  class Pathel {
    float i;
    float j;
    float e;
  }

  Pathel[] els;
  int L=128;
  int q = 0;
  Path() {
    els = new Pathel[L];
    for(int i=0; i<L; i++) {
      els[i] = new Pathel();
    }
  }

  void add(float ii, float jj, float ee) {
    els[q].i = ii;
    els[q].j = jj;
    els[q].e = ee;
    q = (q+1)%L;
  }

  void draw() {
    float t = millis()/5000.0;
    noFill();
    strokeWeight(4);
    beginShape(LINES);
    for(int i=0; i < L; i++) {
      int ei = (i+(q/2)*2)%L;
      stroke(128,255,128,128.0*(i)/L);
      vertex(SW*(A*(els[ei].i+0.5)+B*(els[ei].j+0.5))+TX, SH*(-B*(els[ei].i+0.5)+A*(els[ei].j+0.5)-SE*els[ei].e)+TY);
    }
    endShape();
    noStroke();
  }
}

Path pathy;

class Char {
  float vi;
  float vj;
  float ve;
  float fi;
  float fj;
  int i;
  int j;
  float fe;

  boolean fresh;
  void refresh() {
    fresh = true; 
  }

  void update(float elevation) {
    if(!fresh) {
      return;
    }
    fresh = false;


    float h = 0.1;
    float k = 0.975;

    fi += h*vi;
    fj += h*vj;
    fe += h*ve;
    if(fe < elevation) {
      //fe = elevation;
      fe -=2*h*ve;
      ve *= -0.125;
      fi -=2*h*vi;
      fj -=2*h*vj;
      vi *= -0.5;
      vj *= -0.5;
    }
    vi *= k;
    vj *= k;
    ve *= k;

    if(fe < elevation + 0.25) {
      ve += 0.75*(elevation-fe+0.25);
      ve *= k*k;
    }  
    ve -= 0.05;


    i = (int)(fi+0.5);
    j = (int)(fj+0.5);

    float gk = 0.25;
    if(i > 0 && i < N-1 && j >= 0 && j <= N-1 && fe < elevation + 0.5) {
      vi -= gk*(max(CEL,tunits[(N-i-2)*N+j].elevation) - max(CEL,tunits[(N-i)*N+j].elevation));
    }
    if(i >= 0 && i <= N-1 && j > 0 && j < N-1 && fe < elevation + 0.5) {
      vj += gk*(max(CEL,tunits[(N-i-1)*N+j-1].elevation) - max(CEL,tunits[(N-i-1)*N+j+1].elevation));
    }

    if(i < 0 || i > N-1) { 
      teleport(); 
    }
    if(j < 0 || j > N-1) { 
      teleport(); 
    }

    pathy.add(fi,fj,fe);
  }

  void draw(float elevation) {
    float x = SW*(A*fi+B*fj)+TX;
    float y = SH*(-B*fi+A*fj)+TY;
    fill(0,128,0,16);
    ellipse(x+SW*(A+B)/2,y+SH*(-B+A)/2-SH*SE*elevation,2*SW/3,2*SH/3);
    fill(0,64,0,32);
    ellipse(x+SW*(A+B)/2,y+SH*(-B+A)/2-SH*SE*elevation,SW/3,SH/3);
    fill(0,255,0,16);
    ellipse(x+SW*(A+B)/2,y+SH*(-B+A)/2-SH*SE*fe+2,SW,SH);
    fill(0,255,0);
    ellipse(x+SW*(A+B)/2,y+SH*(-B+A)/2-SH*SE*fe+2,SW/2,SH/2);
    fill(0,0,0);
    ellipse(x+SW*(A+B)/2,y+SH*(-B+A)/2-SH*SE*fe,SW/2,SH/2);
    fill(255,255,255);
    ellipse(x+SW*(A+B)/2,y+SH*(-B+A)/2-SH*SE*fe,SW/8,SH/8);
    noStroke();
  }
}

Char hero;

class Tunit {
  int i;
  int j;
  float nativeElevation;
  float elevation;

  void draw() {
    boolean heroic = i == hero.i && j == hero.j;
    float t = millis()/5000.0;
    float b = min(1.0,2*noise(i/20.0+t,j/20.0+t,t/50.0));

    if(heroic && painting) {
      elevation += 0.5*(hero.fe-elevation-0.10);
    } else {
      elevation += 0.00125*(nativeElevation-elevation); 
    }

    if(true) {
      float x = SW*(A*i+B*j)+TX;
      float y = SH*(-B*i+A*j-SE*elevation)+TY;
      beginShape(QUADS);

      float hb = heroic?0.75:1.0;

      fill(hb*b*245,b*245,hb*b*245);
      vertex(x,y);
      vertex(x+SW*A,y+SH*(-B));
      vertex(x+SW*(A+B),y+SH*(-B+A));
      vertex(x+SW*B,y+SH*(A));


      fill(hb*b*180,b*180,hb*b*180);
      vertex(x,y);
      vertex(x+SW*B,y+SH*(A));
      vertex(x+SW*B,y+SH*(A+SE*(elevation+1)));
      vertex(x,y+SH*SE*(elevation+1));


      fill(hb*b*120,b*120,hb*b*120);
      vertex(x+SW*(A+B),y+SH*(-B+A));
      vertex(x+SW*B,y+SH*(A));
      vertex(x+SW*B,y+SH*(A+SE*(elevation+1)));
      vertex(x+SW*(A+B),y+SH*(-B+A+SE*(elevation+1)));
      endShape();
    }

    if(elevation < CEL) {
      float x = SW*(A*i+B*j)+TX;
      float y = SH*(-B*i+A*j-SE*CEL)+TY;
      beginShape(QUADS);

      fill(b*60,b*100,b*120,128);
      vertex(x,y);
      vertex(x+SW*A,y+SH*(-B));
      vertex(x+SW*(A+B),y+SH*(-B+A));
      vertex(x+SW*B,y+SH*(A));

      endShape(QUADS);

    }

    float ec = max(elevation,CEL);

    if(heroic) {
      lastHeroicTunit = this;
      hero.update(ec);
      hero.draw(ec);
    }

  }
}
void setupTerrain() {
  float sk = 4.0/N;
  if(tunits == null) {
    tunits = new Tunit[N*N];
  }

  float p = millis()/10000.0;
  for(int i = 0; i < N; i++) {
    for(int j = 0; j < N; j++) {
      Tunit t = new Tunit();
      t.i = N-i-1;
      t.j = j;
      t.elevation = pow(2.0*abs(noise(sk*i,sk*j,p)-0.5)-0.75,3);
      t.nativeElevation = t.elevation;
      tunits[i*N+j] = t;
    }
  }
  lastHeroicTunit = tunits[0];
  teleport();
}

void teleport() {
  for(int q = 0; q < 100; q++) {
    int qi = (int)random(N);
    int qj = (int)random(N);

    if(tunits[qi*N+qj].elevation > CEL) {
      hero.fi = tunits[qi*N+qj].i;
      hero.fj = tunits[qi*N+qj].j;
      hero.vj *= 0.5;
      hero.vi *= 0.5;
      hero.ve = 0;

      hero.refresh();
      hero.update(tunits[qi*N+qj].elevation);
      break; 
    }
  }
  refocus(1.0);
}

void refocus(float k) {
  TX += k*(-SW*(A*hero.fi+B*hero.fj)+width/2-TX);
  TY += k*(-SH*(-B*hero.fi+A*hero.fj-SE*hero.fe)+height/2-TY);
}

void setup() {
  size(640,480);
  hero = new Char();
  pathy = new Path();
  SW = (width/W)/N;
  SH = (height/H)/N;
  SE = (height/E);
  SEL = (height/EL);
  setupTerrain();
}

void draw() {
  refocus(0.1);
  scroll();

  background(102);

  noStroke();
  for(int idx = 0; idx < N*N; idx++) {
    tunits[idx].draw();
  }
  pathy.draw();
  hero.refresh();
}

void mousePressed() {
  if(mouseButton == LEFT) {
    painting = true; 
  }
  if(mouseButton == RIGHT) {
    scrolling = true;
    scrolltx = mouseX;
    scrollty = mouseY;
  }
} 

void mouseReleased() {
  if(mouseButton == LEFT) {
    painting = false; 
  }
  if(mouseButton == RIGHT) {
    scrolling = false; 
  }
}

void scroll() {
  if(scrolling) {
    float k = 0.1;
    TX -= k*(mouseX-scrolltx);
    TY -= k*(mouseY-scrollty);
  }
}

void keyPressed() {
  if(key == 'r') {
    setupTerrain();
  } 

  if(key == ' ' && hero.fe < 0.33) { 
    hero.ve++; 
  }
  if(key == 'w') { 
    hero.vi++; 
  }
  if(key == 'a') { 
    hero.vj--; 
  }
  if(key == 's') { 
    hero.vi--; 
  }
  if(key == 'd') { 
    hero.vj++; 
  }

}


















